#!/tmp/.ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZpErLZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZperl/bin/perl
use strict;
use PureMessage::Version qw($PREFIX);
use HTML::Template;
use PureMessage::Manager::Request qw($REQUEST %in);
use PureMessage::Manager::Response::Policy::Spam;
use PureMessage::Manager::Widget;
use PureMessage::Manager::Widget::Grid;
use PureMessage::Manager::Lang;
use PureMessage::Manager::Utils qw(html_escape);
use PureMessage::Policy::Spam::Rules;
use PureMessage::Policy::Spam::Config;
use PureMessage::Manager::Config qw(%config);
use HTML::Entities;

# get a relative to prefix path for our shared files
for (qw(userdatadir datadir)) {
    $config{$_} = $PureMessage::Policy::Spam::Config::config{$_};
    $config{$_} =~ s#^\Q$PREFIX\E/?##;
}

# What files should we offer to publish?
my @PUBLISH;
{
    (my $d  = $config{datadir})     =~ s#^\Q$PREFIX\E/##;
    (my $dd = $config{userdatadir}) =~ s#^\Q$PREFIX\E/##;
    @PUBLISH = ("$d/lock", map "$dd/$_", qw(
	re.rules
	db.pdelta
	db.force
    ));
}

# what plugins should not appear in the drop down search list
my $TABLE_WIDTH = '100%';

my $PART_JAVASCRIPT = <<EOF;
<SCRIPT LANGUAGE='JavaScript'>
  function partPop() {
    sList=window.open('pol/mgr_pol_message_parts','Parts','toolbar=no,location=no,directories=no,status=no,menubar=no,scrollbars=yes,resizable=no,width=550,height=550,top=100,left=100');
  }
</SCRIPT>  
EOF

# we absolutely require these for the manager. we require name
# also but it will always be there anyhow so we don't worry
# about it
my @REQUIRED_MANAGER_FIELDS = qw(Enabled Desc w ProbabilityDelta);
my %EDITABLE_MANAGER_FIELDS = map { $_ => 1 } qw(Enabled Desc Part Test);


my ($my_rules, $static_rules);
my $WIDGET = new PureMessage::Manager::Widget;

my $r = $REQUEST->get_response( 'PureMessage::Manager::Response::Policy::Spam',
                                'stderr_to_stdout' =>1
                              );

if($in{start_filter_by}) {
    $r->add_help(text => "Help", tab => "pol", page => "mgr_pol_spam_rules" );
}
else {
    $r->add_help(text => "Help", tab => "pol", page => "mgr_pol_spam_rules_create" );
}

$r->title(text('spam_rules_title'));

$r->{debug} = 0; # debug will break part popup

my %TBattr;
my ($filter, $show_enabled, $show_only, $show_plugin);
if ($in{start_filter_by}) {
    $filter         = $in{filter_by};
    $show_enabled   = $in{show_enabled};
    $show_only      = $in{show_only};
    $show_plugin    = $in{show_plugin};
}
else {
    $filter         = $in{current_filter};
    $show_enabled   = $in{current_show_enabled};
    $show_only      = $in{current_show_only};
    $show_plugin    = $in{current_show_plugin};
}

my ($rules, @rules, @groups);

eval {
    $rules = PureMessage::Policy::Spam::Rules->new;
    init_rules_data();
};
if ($@) {
    $r->raise_error("$@");
    $r->send();
    exit;
}

# Publications widget
eval {
    require PureMessage::Manager::Response::ServerGroups::Publications;
    $r->print(
        PureMessage::Manager::Response::ServerGroups::Publications::Publish(
            \@PUBLISH,
	    "AntiSpam Rules",
	)
	. "<br>"
    );
};

my $edit = $in{start_filter_by} ? '' : $in{edit};
my $rule_widget = get_rule_widget($edit);

if ($rule_widget->read_input($REQUEST)) {
    my $data = $rule_widget->result_data;
    if ($data->{save}) {
	# try to save this data then
	$rule_widget->set('state',{ edit => undef });
	my $be;
	my ($rule,@junk) = grep { $_->{_name} eq $edit} @rules;
	if ($rule) {
	    my %data = %$rule;
	    my %editable = getEditableFields($rule);
	    foreach my $k ( getrulekeys()) {
		$data{umkrulekey($k)} = $in{$k} if $editable{umkrulekey($k)};
	    }

	    $data{Name} = $rule->{_name};
	    delete($data{c_enabled}) if $editable{Enabled};

	    # if this isn't a user_rule right then it's a static rule
	    # soon to become an overriden rule when we are done with it.
	    # set key so that the validation knows this too
	    $data{c_overridden} = 1 if !$rule->{c_user};

	    if (!validData(\%data, $rule)) {
		$TBattr{reshow} =1;
		$edit = $rule->{_name};
		$rule_widget->set('state',{ edit => $edit });
	    }
	    else {
		eval {
		    $rules->editrule($data{Name},\%data);
		    $rules->save;
		};
		if ($@) {
		    $r->raise_error(text('error_saving_rules',$@));
		    #  and reshow this edit
		    $TBattr{reshow} = 1;
		    $edit = $rule->{_name};
		    $rule_widget->set('state',{ edit => $edit});
		}
		else {
		    init_rules_data();     
		    $edit = undef;
		    after_save();
		}    
	    }

	}
	else {
            $r->raise_error(text('error_rules_no_such_rule',encode_entities($edit,'&<>"\/\'')));
	}
    }
    elsif ($data->{cancel} or $data->{new_page}) {
	$rule_widget->set('state',{ edit => undef });
	$edit = undef;
    }
    elsif ($data->{new}) {
	$TBattr{show_new} = 1;

	# If we were filtering before, ensure that our new rule will be
	# visible when the user clicks "SAVE".
	$show_plugin = "SITE" if $show_plugin and $show_plugin ne "all";
	$filter      = undef  if $filter;

	$rule_widget->set('state',{ edit => undef });
	$edit = undef;
    }
    elsif ($data->{save_new}) {
	my ($rule,@junk) = grep { $_->{_name} eq $in{new_name}} @rules;
	if ($rule) {
	    # that rule exists
	    $r->raise_error(text('error_name_in_use',$in{new_name}));
	    $TBattr{show_new} = 1;
	}
	else {
	    my %data;
	    $data{w} = $in{mkrulekey('w')};
	    $data{Desc} = $in{mkrulekey('Desc')};
	    $data{ProbabilityDelta} = $in{mkrulekey('ProbabilityDelta')};
	    if ($in{mkrulekey('Part')} !~ /,/) {
		$data{Part} = $in{mkrulekey('Part')};
	    }
	    else {
		my @part = map { s/ //; $_ } split(',',$in{mkrulekey('Part')});
		$data{Part} = \@part;
	    }
	    $data{Enabled} = $in{mkrulekey('Enabled')};
	    $data{Test} = $in{mkrulekey('Test')};
	    $data{Name} = $in{new_name};
	    if (validData(\%data)) {
		$rules->addrule($data{Name},\%data);
		$rules->save();
		init_rules_data();
		after_save(); 
	    }
	    else {
		$TBattr{show_new} = 1;
	    }
	}
	$rule_widget->set('state',{ edit => undef });
	$edit = undef;
	
    }
    elsif ($data->{delete} or $data->{reset}) {
	my ($rule,@junk) = grep { $_->{_name} eq $edit} @rules;
	if ($rule) {
	    $rules->delrule($edit) if $data->{delete};
	    $rules->revertrule($edit) if $data->{reset};
	    $rules->save();
	    init_rules_data();
	    after_save();

	}
	else {
	    $r->raise_error(text('error_rules_no_such_rule',$edit));
	}
	$rule_widget->set('state',{ edit => undef });
	$edit = undef;
    }
}

$r->add_part("<form action=rules.cgi method=post name=rule_edit>",'start_form');
$r->add_part("<table width=600><tr><td>");

$r->print("<br>");
$r->add_part(_format_filter(),'filter');
$r->add_part($rule_widget,'rules');
$r->add_part("</table>");
$r->add_part("</form>",'end_form');

$r->send;

# subs
#
#

sub validData {
    my %data = %{shift(@_)};
    my $rule = shift;
    my %check;

    %check = getEditableFields($rule);

    $r->raise_error('error_rules_bad_rule_name')
	unless $data{Name} =~ /^[A-Za-z_][A-Za-z0-9_]*$/;
    $r->raise_error('error_rules_score_bad_form_score')
	if $check{w}
	    and $data{w} !~ /^(\+|-)?\d+(\.\d+)?$/;
    $r->raise_error('error_rules_score_bad_form_probability')
	if $check{ProbabilityDelta}
	    and $data{ProbabilityDelta} !~/^(\+|-)?\d+(\.\d+)?$/;
    $r->raise_error('error_rules_empty_description')
	if $check{Desc} and $data{Desc} !~ /^.+$/;
    $r->raise_error('error_rules_test_bad_form')
	if $check{Test}
	    and exists $data{Test}
	    and $data{Test} !~ m{^(\s*!\s*)?/.*/[gimosx]*$};
    
    if ($check{Part}) {
	$r->raise_error('error_rules_empty_part')
	    unless defined $data{Part} && length $data{Part};
	$r->raise_error('error_rules_bad_part', $data{Part})
	    unless $data{Part} =~ /^[\w\-:]+$/;
    }

    return $r->error() ? 0 : 1;
}

sub init_rules_data  {
    # Get a list of valid groups
    @groups = grep $rules->getgroup($_), $r->get_valid_backends;

    my @r;

    # If there is a filter now is the time to apply it
    if ($show_plugin and $show_plugin ne 'all') {
	@r = $rules->rulenames($show_plugin);
    }
    # Otherwise, grab all the rule names
    else {
	# I'm torn between showing *all* the rules, or just the rules that
	# belong to groups visible in the manager. Currently I favour showing
	# all the rules, since they do in fact appear in the reports and
	# message headers.
	@r = $rules->rulenames;
    }

    @rules = ();
    foreach (@r) {
	my $data = $rules->getrule($_);

	# Apply various filters
	if ($filter) {
	    next unless /\Q$filter\E/i;
	}
	if ($show_plugin ne "SITE") {
	    if ($show_only eq 'over') {
		next unless $data->{c_overridden};
	    }
	    if ($show_only eq 'static') {
		next if $data->{c_user} or $data->{c_overridden};
	    }
	}
	if ($show_enabled =~ /^(?:yes|no)$/) {
	    my $c_enabled = $show_enabled eq 'yes' ? 1 : 0;
	    next if $data->{Enabled} != $c_enabled;
	}

	# Transform the definition into the old format.
	$data->{_name} = $_;
	$data->{c_enabled} = delete $data->{Enabled};
	$data->{Enabled} = $data->{c_overridden} ?
			   $data->{c_enabled}    ? "yes" : "no"
						 : "__DEFAULT__";
	push (@rules, $data);
    }
}

sub make_edit {
    my ($name,$value,$size) = @_;
    $value = html_escape($value);
    $size = $size || '10';
    return "<input class=stretch type=type name=$name size=$size value='$value'>\n";
}

sub TABLE {
    my @c = @_;
    my $data = "<table cellpadding=0 cellspacing=0 width='100%'>\n";
    foreach (@c) {
	$data .= ROW(@$_);
    }
    return "$data</table>\n";
}

sub ROW {
    my @c = @_;
    my $data = "<tr>\n";
    my $i = 0;
    my $size;
    foreach (@c) {
	$size = $i ? 'width="100%"' : '';
	$data .= "<td nowrap $size>$_</td>\n";
	$i++;
    }
    return "$data</tr>\n"
}

sub get_newedit_row {
    my $edit = shift;
    my %data;
    $data{LOCATION_ICON} = '';
    $data{RULE_NAME} = make_edit('new_name',$in{new_name},25);
    my @DEF = (
	["Desc: ",
	make_edit(
	    mkrulekey('Desc'),
	    $in{mkrulekey('Desc')},
	    30,
	)],
	["Part: ",
	make_edit(
	    mkrulekey('Part'),
	    $in{mkrulekey('Part')},
	    8,
	) . $PART_JAVASCRIPT,
	],
	['Test: ',
	make_edit(
	    mkrulekey('Test'),
	    $in{mkrulekey('Test')},
	    25,
	)],
    );
    $data{W} = make_edit(
	mkrulekey('w'),
	defined($in{mkrulekey('w')}) ? $in{mkrulekey('w')} : 0,
	6,
    ); 
    $data{PROBABILITY} = make_edit(
	mkrulekey('ProbabilityDelta'),
	defined($in{mkrulekey('ProbabilityDelta')})
	    ? $in{mkrulekey('ProbabilityDelta')}
	    : 0,
	6
    );
    $data{ENABLED} = $WIDGET->option_list(
	mkrulekey('Enabled'),
	{
	    '__DEFAULT__' => 'Auto',
	    'no' => 'Disabled',
	    'yes' => 'Enabled'
	},
	defined($in{mkrulekey('Enabled')})
	    ? $in{mkrulekey('Enabled')}
	    : 'yes',
    );
   
    return (TABLE([$data{ENABLED}]),$data{RULE_NAME},
				    TABLE(@DEF),
				    $data{W},
				    $data{PROBABILITY});
    
}

sub getEditableFields { 
    # A rule edit is a mix of what data is available in the rule
    # definition, what data we are allowed to edit for this
    # feature group as defined in the override key, and what data we have
    # chosen to expose editing for via the manager
    my $rule = shift;
    my %editable;

    if (defined $rule) {
	foreach (keys %{$rule->{override}}) {
	    $editable{$_} = 1
		if exists($rule->{$_}) and $EDITABLE_MANAGER_FIELDS{$_};
	}
    }
    else {
	%editable = %EDITABLE_MANAGER_FIELDS;
    }

    # these are always editable but aren't listed in override
    $editable{$_} = 1 foreach qw(w ProbabilityDelta); 
    return %editable;
}

sub mkrulekey {
  my $key = shift;
  return '_rl_' . $key;
}

sub umkrulekey {
  my $key = shift;
  $key =~ s/^_rl_//;
  return $key;
}

sub getrulekeys {
 my @keys = grep { /^_rl_/} keys %in;
 return @keys;
}

sub get_edit_row {
    my $rule = shift;
    my $reshow = shift; # is a reshowing of an edit?

    unless ($rule) {
	$r->raise_error(text('error_rules_no_such_rule', $rule));
	return;
    }

    my %editable = getEditableFields($rule);

    # fixup the rule data we have so that if a required field
    # doesn't exist it's got a nice default
    for my $field (@REQUIRED_MANAGER_FIELDS) {
	$rule->{$field} = ' ' unless exists $rule->{$field};
    }

    my %data = (
	LOCATION_ICON   => get_rule_icon($rule),
	RULE_NAME       => $rule->{_name},
    );

    if ($editable{Enabled}) {
	$data{ENABLED} = $WIDGET->option_list(
	    mkrulekey('Enabled'),
	    {
		'__DEFAULT__' => 'Auto',
		'no' => 'Disabled',
		'yes' => 'Enabled',
	    },
	    $reshow ? $in{enabled} : $rule->{Enabled},
	);
    }
    else {
	$data{ENABLED} = $data{LOCATION_ICON};
    }
    delete $rule->{Enabled};

    # Weight and ProbabilityDelta are always editable
    foreach my $k (qw(w ProbabilityDelta)) {
	$data{uc($k)} = make_edit(
	    mkrulekey($k),
	    $reshow ? $in{mkrulekey($k)} : $rule->{$k},
	    6,
	);
	delete $rule->{$k};
    }

    # There are two types of rules: site rules and other. If this rule is a
    # site rule, we can edit qw(Enabled Desc Part Test). Otherwise we can only
    # edit Enabled, so just return what we have right now.
    unless ($rules->memberOf($rule->{_name}, "SITE")) {
	return (TABLE([$data{ENABLED}]), $data{RULE_NAME},
					 formatted_desc($rule),
					 $data{W},
					 $data{PROBABILITYDELTA});
    }

    # Write the definition column.
    my @DEF;
    if ($editable{Desc}) {
	push @DEF, ["Desc: ", make_edit(
	    mkrulekey('Desc'),
	    $reshow ? $in{mkrulekey('Desc')} : $rule->{Desc},
	    30,
	)];
    }
    else {
	push @DEF, ["Desc: ", html_escape($rule->{Desc})];
    }
    my $info_part = delete $rule->{Part};
    if ($editable{Part}) {
	push @DEF, [
	    "Part: ",
	    make_edit(
		mkrulekey('Part'),
		$reshow ? $in{mkrulekey('Part')} : $info_part,
		8
	    ) . $PART_JAVASCRIPT,
	];
    }
    else {
	push @DEF, ["Part: ", $info_part];
    }
    if (exists $rule->{Test}) {
	if ($editable{Test}) {
	    push @DEF, ["Test: ", make_edit(
		mkrulekey("Test"),
		$reshow ? $in{mkrulekey("Test")} : $rule->{Test},
		25,
	    )];
	}
	else {
	    push @DEF, ["Test: ", html_escape($rule->{Test})];
	}
	delete $rule->{Test};
    }

    # Get the keys that are left and show them to the user, only ucfirst
    # keys are user visible
    my @other_keys = grep { /^[A-Z]/ } keys %{$rule};
    foreach my $k (@other_keys) {
	next if $k eq "Desc" or $k eq "LongDesc";
	if ($editable{$k}) {
	    push @DEF, [
		"$k: ",
		make_edit(
		    mkrulekey($k),
		    $reshow ? $in{mkrulekey($k)} : $rule->{$k},
		    25,
		),
	    ];
	}
	else {
	    push @DEF, ["$k: ", html_escape($rule->{$k})];
	}
	delete $rule->{$k};
    }
    my @r= (TABLE([$data{ENABLED}]), $data{RULE_NAME},
				     TABLE(@DEF),
				     $data{W},
				     $data{PROBABILITYDELTA});
    @r;
}


sub get_rule_icon {
    my ($rule) = @_;
    my $icon = 'rule_';
    my $text;
    $icon .= 'o',$text = 'Override' if $rule->{c_overridden};
    $icon .= 's',$text = 'Site' if $rule->{c_user} and !$rule->{c_overridden};
    $icon .= 'd',$text = 'Default' if !$rule->{c_user} and !$rule->{c_overridden};
    $icon .= $rule->{c_enabled} ? '_enabled' : '_disabled';
    return button({ type => 'image', alt => $text, icon => $icon});
}

sub gen_rule_data {
    my ($start,$end) = @_;
    my @data;
    if (defined $start and defined $end) {
	for my $i ($start .. $#rules) {
	    last if $i > $end;

	    my $rule = $rules[$i];
	    my $icon = get_rule_icon($rule);

	    # the generated text is 'slaved' to the edit hidden
	    # field generated by the grids state param. If we don't
	    # do this we will create a hidden field for every text
	    # button
	    my @edit;
	    if ($edit eq $rule->{_name} and
		@edit = get_edit_row($rule, $TBattr{reshow}))
	    {
		push @data, [@edit];
		$TBattr{save} = 1;
		if ($rule->{c_overridden}) {
		    $TBattr{reset} = 1;
		}
		elsif ($rule->{c_user} and !$rule->{c_overriden}) {
		    $TBattr{delete} = 1;
		}
	    }
	    else {
		my $test;
		my $href;
		if (!$edit and !$TBattr{show_new}) {
		    $href = button({
			    type => 'text',
			    name => 'edit',
			    slave => 'edit',
			    value => $rule->{_name},
			    text => clipLongStr($rule->{_name}, 30),
		    });
		}
		else {
		    $href = clipLongStr($rule->{_name}, 30);
		}
		push(@data,[$icon, $href,
			    TABLE([formatted_desc($rule)]),
			    $rule->{w},
			    $rule->{ProbabilityDelta}]);
	    }
	}
	if ($TBattr{show_new}) {
	    my $index = $#data >= 0 ? $#data : 0;
	    $data[$index] = [get_newedit_row()]; # put our row as the last one
	}
	return \@data;
    }
    else {
	if ($TBattr{show_new}) {
	    # then we have to ensure that there is at least one row for us
	    return $#rules >= 0 ? $#rules : 0;
	}
	else {
	    return $#rules;
	}
    }
}


sub get_rule_widget {
    my $rule_widget = $WIDGET->widget(
	CLASS => 'PureMessage::Manager::Widget::Grid',
	name => 'rules_grid',
	data => \&gen_rule_data,
	page => 0,
	selectable => 0,
	page_size => 20,
	paging => 1,
	columns => 5,
	show_count => 0,
	toolbar => sub {
	    my @buttons;
	    
	    if (!$TBattr{save} and !$TBattr{show_new}) {
		push @buttons, {
		    type => 'image',
		    icon => 'add',
		    name => 'new',
		    text => 'New',
		};
	    }

	    if ($TBattr{reset}) {
	        push @buttons, {
		    type => 'image',
		    icon => 'clear',
		    name => 'reset',
		    text => 'Revert',
		};
	    }

	    if ($TBattr{delete}) {
	        push @buttons, {
		    type => 'image',
		    icon => 'delete',
		    name => 'delete',
		    text => 'Delete'
		};
	    }

	    if ($TBattr{show_new}) {
		push(@buttons,
		    {
			type => 'image',
			icon => 'save',
			name => 'save_new',
			text => 'Save'
		    },
		    {
			type => 'image', 
			icon => 'cancel',
			name => 'cancel',
			text => 'Cancel'
		    },
		);
	    }

	    if ($TBattr{save}) {
	        push (@buttons,
		    {
			type => 'image',
			icon => 'save',
			name => 'save',
			text => 'Save'
		    },
		    {
			type => 'image', 
			icon => 'cancel',
			name => 'cancel',
			text => 'Cancel'
		    },
		);
	    }

	    return $WIDGET->widget(
		CLASS => 'PureMessage::Manager::Widget::ToolBar',
		buttons => [\@buttons],
		state => {
                    current_show_plugin => encode_entities($show_plugin,'&<>"\/\''),
                    current_show_only => encode_entities($show_only,'&<>"\/\''),
                    current_show_enabled => encode_entities($show_enabled,'&<>"\/\''),
                    current_filter => encode_entities($filter,'&<>"\/\'')
		},
	    );
	},
        state => { edit => encode_entities($edit,'&<>"\/\'')},
	width => $TABLE_WIDTH,
	padding => '<BR> <BR>',
	ddata => [
	    { 
		width => 10,
		description => '',
		extra => 'nowrap valign=top',
	    },
	    { 
		width => 100,
		description => 'Rule Name',
		extra => 'nowrap valign=top',
	    },
	    { 
		width => 250,
		description => 'Rule Definition',
		extra => 'nowrap',
	    }, 
	    { 
		width => 70,
		description => 'Weight',
		extra => 'nowrap valign=top',
	    }, 
	    { 
		width => 60,
		description => 'Probability Adjust %',
		extra => 'nowrap valign=top',
	    }, 
	],
    );   

    return $rule_widget;
}


sub _format_filter {
  my $result;
  my $filter_value = encode_entities($in{filter_by},'&<>"\/\'');
  my @buttons = ({ type => 'hidden',
                   name => 'displayed_filter_by',
                   value => $filter_value,
                 },
                 { type => 'custom',
                   text => "<input type=text size=10 name=filter_by value='$filter_value'>"
                 },
                 { type => 'image',
                   name => 'start_filter_by',
                   icon => 'filter',
                   text => 'Filter',
                 },
                );

  my @bd = buttons(\@buttons);
  my $filter = join('&nbsp;',@bd);
  my $DATA = { all => 'All Rules',
               static => 'Default Rules',
               'over' => 'Overridden Rules (O)',
             };

  my @order = qw(all static over);
  my %PDATA = (all => "All Groups");
  my @all_plugins = qw(all);

  foreach (@groups) {
    my $info = $rules->getgroup($_);
    $PDATA{$_} = $info->{Desc};
    push @all_plugins, $_;
  }

  my $PLUGINS = $WIDGET->option_list('show_plugin',\%PDATA,$show_plugin,'',\@all_plugins);
  my $SHOW_ENABLED = $WIDGET->option_list('show_enabled',{ 'any' => 'Any State',yes => 'Enabled', no => 'Disabled'},$show_enabled,'',['any','yes','no']);
  my $DROP_DOWN = $WIDGET->option_list('show_only',$DATA,$show_only,'',\@order);
  $filter =<<EOF;
        $bd[0]
        <table cellpadding=2 cellspacing=0>
        <tr>
          <td>Feature Group</td>
          <td>$PLUGINS</td>
          <td>Show</td>
          <td>$DROP_DOWN</td>
          <td>$SHOW_ENABLED</td>
          <td>$bd[1]</td>
          <td>$bd[2]</td>
        </tr>
        </table>
EOF

  my $data;
  $data = "<table cellpadding=0 cellspacing=0 width='$TABLE_WIDTH'>";
  $data .= "<tr><td align=right valign=top >\n";
  $data .= $WIDGET->widget( CLASS => 'PureMessage::Manager::Widget::PaneBox',
                           content => $filter,
                           width=> $TABLE_WIDTH,
                           suppress_bottom_border => 1,
                           td_extra => 'align=right valign=top nowrap',
                         )->output;
  $data .= "</td></tr>\n";
  return $data;                          
}


sub clipLongStr {
  my ($string,$len) = @_;
  return $string if length($string) <= $len;
  my $newString;
  $newString = substr($string,0,$len);
  return "$newString...";
}

sub after_save {
    # allow PM::Service::Config to not exist, like
    # for instance, if they have an older PM-Utils
    # without it.
    eval {
        require PureMessage::Service::Config;
        PureMessage::Service::Config::stale('milter');
    };
}

sub formatted_desc {
    my $rule = shift;
    my @lines = "<table>\n";
    for (split /[\r\n]+/, $rule->{LongDesc}) {
	my ($lhs, $rhs) = split /:\s+/, $_, 2;
	$_ = html_escape($_) for ($lhs, $rhs);
	push @lines, "<tr><td valign='top'>$lhs:</td><td>$rhs</td></tr>\n";
    }
    push @lines, "</table>\n";
    return join('', @lines);
}
